package networking

import (
	"encoding/json"
	"fmt"
	"net"
)

type Server[T any] struct {
	listener net.Listener
	C        chan Connection[T]
}

type Client[T any] struct {
	conn net.Conn
}

type Connection[T any] struct {
	Conn    net.Conn
	encoder *json.Encoder
	decoder *json.Decoder
}

func NewServer[T any](port int) Server[T]{

    listener, err := net.Listen("tcp", fmt.Sprintf("0.0.0.0:%d", port))
	if err != nil {
		panic("Server could not bind to address")
	}
   return Server[T]{
        listener:listener,
        C: make(chan Connection[T]),
    } 
}

func (s *Server[T]) ListenLoop() {

	for {
		listenerConn, err := s.listener.Accept()
		if err != nil {
			panic("Server could not accept connection")
		}
        conn := NewConnection[T](listenerConn)
		s.C <- conn
	}
}

func (c Client[T]) Connect() Connection[T] {
	dialConn, err := net.Dial("tcp", "localhost:8080")
	if err != nil {
		panic("Could not open connection to server")
	}
	return NewConnection[T](dialConn)
}

func NewConnection[T any](netConn net.Conn) Connection[T] {
	return Connection[T]{
		Conn:    netConn,
		encoder: json.NewEncoder(netConn),
		decoder: json.NewDecoder(netConn),
	}
}

func (jc Connection[T]) Send(obj T) {
	if err := jc.encoder.Encode(&obj); err != nil {
		panic("Failed encoding data or sending it to connection")
	}
}

func (jc Connection[T]) Receive() T {
	var obj T
	if err := jc.decoder.Decode(&obj); err != nil {
		panic("Failed decoding data or reading it from connection")
	}
	return obj
}
